@@ -26,19 +26,15 @@ class HTTPConnection(ConnectionInterface):
     def __init__(
         self,
         origin: Origin,
-        ssl_context: ssl.SSLContext = None,
-        keepalive_expiry: float = None,
+        ssl_context: Optional[ssl.SSLContext] = None,
+        keepalive_expiry: Optional[float] = None,
         http1: bool = True,
         http2: bool = False,
         retries: int = 0,
-        local_address: str = None,
-        uds: str = None,
-        network_backend: NetworkBackend = None,
+        local_address: Optional[str] = None,
+        uds: Optional[str] = None,
+        network_backend: Optional[NetworkBackend] = None,
     ) -> None:
-        ssl_context = default_ssl_context() if ssl_context is None else ssl_context
-        alpn_protocols = ["http/1.1", "h2"] if http2 else ["http/1.1"]
-        ssl_context.set_alpn_protocols(alpn_protocols)
-
         self._origin = origin
         self._ssl_context = ssl_context
         self._keepalive_expiry = keepalive_expiry
@@ -52,6 +48,7 @@ class HTTPConnection(ConnectionInterface):
             SyncBackend() if network_backend is None else network_backend
         )
         self._connection: Optional[ConnectionInterface] = None
+        self._connect_failed: bool = False
         self._request_lock = Lock()
 
     def handle_request(self, request: Request) -> Response:
@@ -62,27 +59,31 @@ class HTTPConnection(ConnectionInterface):
 
         with self._request_lock:
             if self._connection is None:
-                stream = self._connect(request)
+                try:
+                    stream = self._connect(request)
 
-                ssl_object = stream.get_extra_info("ssl_object")
-                http2_negotiated = (
-                    ssl_object is not None
-                    and ssl_object.selected_alpn_protocol() == "h2"
-                )
-                if http2_negotiated or (self._http2 and not self._http1):
-                    from .http2 import HTTP2Connection
+                    ssl_object = stream.get_extra_info("ssl_object")
+                    http2_negotiated = (
+                        ssl_object is not None
+                        and ssl_object.selected_alpn_protocol() == "h2"
+                    )
+                    if http2_negotiated or (self._http2 and not self._http1):
+                        from .http2 import HTTP2Connection
 
-                    self._connection = HTTP2Connection(
-                        origin=self._origin,
-                        stream=stream,
-                        keepalive_expiry=self._keepalive_expiry,
-                    )
-                else:
-                    self._connection = HTTP11Connection(
-                        origin=self._origin,
-                        stream=stream,
-                        keepalive_expiry=self._keepalive_expiry,
-                    )
+                        self._connection = HTTP2Connection(
+                            origin=self._origin,
+                            stream=stream,
+                            keepalive_expiry=self._keepalive_expiry,
+                        )
+                    else:
+                        self._connection = HTTP11Connection(
+                            origin=self._origin,
+                            stream=stream,
+                            keepalive_expiry=self._keepalive_expiry,
+                        )
+                except Exception as exc:
+                    self._connect_failed = True
+                    raise exc
             elif not self._connection.is_available():
                 raise ConnectionNotAvailable()
 
@@ -132,8 +133,16 @@ class HTTPConnection(ConnectionInterface):
                 break
 
         if self._origin.scheme == b"https":
+            ssl_context = (
+                default_ssl_context()
+                if self._ssl_context is None
+                else self._ssl_context
+            )
+            alpn_protocols = ["http/1.1", "h2"] if self._http2 else ["http/1.1"]
+            ssl_context.set_alpn_protocols(alpn_protocols)
+
             kwargs = {
-                "ssl_context": self._ssl_context,
+                "ssl_context": ssl_context,
                 "server_hostname": self._origin.host.decode("ascii"),
                 "timeout": timeout,
             }
@@ -147,34 +156,39 @@ class HTTPConnection(ConnectionInterface):
 
     def close(self) -> None:
         if self._connection is not None:
-            self._connection.close()
+            with Trace("connection.close", None, {}):
+                self._connection.close()
 
     def is_available(self) -> bool:
         if self._connection is None:
             # If HTTP/2 support is enabled, and the resulting connection could
             # end up as HTTP/2 then we should indicate the connection as being
             # available to service multiple requests.
-            return self._http2 and (self._origin.scheme == b"https" or not self._http1)
+            return (
+                self._http2
+                and (self._origin.scheme == b"https" or not self._http1)
+                and not self._connect_failed
+            )
         return self._connection.is_available()
 
     def has_expired(self) -> bool:
         if self._connection is None:
-            return False
+            return self._connect_failed
         return self._connection.has_expired()
 
     def is_idle(self) -> bool:
         if self._connection is None:
-            return False
+            return self._connect_failed
         return self._connection.is_idle()
 
     def is_closed(self) -> bool:
         if self._connection is None:
-            return False
+            return self._connect_failed
         return self._connection.is_closed()
 
     def info(self) -> str:
         if self._connection is None:
-            return "CONNECTING"
+            return "CONNECTION FAILED" if self._connect_failed else "CONNECTING"
         return self._connection.info()
 
     def __repr__(self) -> str:
@@ -188,8 +202,8 @@ class HTTPConnection(ConnectionInterface):
 
     def __exit__(
         self,
-        exc_type: Type[BaseException] = None,
-        exc_value: BaseException = None,
-        traceback: TracebackType = None,
+        exc_type: Optional[Type[BaseException]] = None,
+        exc_value: Optional[BaseException] = None,
+        traceback: Optional[TracebackType] = None,
     ) -> None:
         self.close()
