@@ -1,9 +1,17 @@
-from typing import List
+import logging
+from typing import List, Optional
 
 import pytest
 import trio as concurrency
 
-from httpcore import AsyncConnectionPool, ConnectError, UnsupportedProtocol
+from httpcore import (
+    AsyncConnectionPool,
+    ConnectError,
+    PoolTimeout,
+    ReadError,
+    UnsupportedProtocol,
+)
+from httpcore.backends.base import AsyncNetworkStream
 from httpcore.backends.mock import AsyncMockBackend
 
 
@@ -153,6 +161,74 @@ async def test_trace_request():
     ]
 
 
+@pytest.mark.anyio
+async def test_debug_request(caplog):
+    """
+    The 'trace' request extension allows for a callback function to inspect the
+    internal events that occur while sending a request.
+    """
+    caplog.set_level(logging.DEBUG)
+
+    network_backend = AsyncMockBackend(
+        [
+            b"HTTP/1.1 200 OK\r\n",
+            b"Content-Type: plain/text\r\n",
+            b"Content-Length: 13\r\n",
+            b"\r\n",
+            b"Hello, world!",
+        ]
+    )
+
+    async with AsyncConnectionPool(network_backend=network_backend) as pool:
+        await pool.request("GET", "http://example.com/")
+
+    assert caplog.record_tuples == [
+        (
+            "httpcore",
+            logging.DEBUG,
+            "connection.connect_tcp.started host='example.com' port=80 local_address=None timeout=None",
+        ),
+        (
+            "httpcore",
+            logging.DEBUG,
+            "connection.connect_tcp.complete return_value=<httpcore.AsyncMockStream>",
+        ),
+        (
+            "httpcore",
+            logging.DEBUG,
+            "http11.send_request_headers.started request=<Request [b'GET']>",
+        ),
+        ("httpcore", logging.DEBUG, "http11.send_request_headers.complete"),
+        (
+            "httpcore",
+            logging.DEBUG,
+            "http11.send_request_body.started request=<Request [b'GET']>",
+        ),
+        ("httpcore", logging.DEBUG, "http11.send_request_body.complete"),
+        (
+            "httpcore",
+            logging.DEBUG,
+            "http11.receive_response_headers.started request=<Request [b'GET']>",
+        ),
+        (
+            "httpcore",
+            logging.DEBUG,
+            "http11.receive_response_headers.complete return_value="
+            "(b'HTTP/1.1', 200, b'OK', [(b'Content-Type', b'plain/text'), (b'Content-Length', b'13')])",
+        ),
+        (
+            "httpcore",
+            logging.DEBUG,
+            "http11.receive_response_body.started request=<Request [b'GET']>",
+        ),
+        ("httpcore", logging.DEBUG, "http11.receive_response_body.complete"),
+        ("httpcore", logging.DEBUG, "http11.response_closed.started"),
+        ("httpcore", logging.DEBUG, "http11.response_closed.complete"),
+        ("httpcore", logging.DEBUG, "connection.close.started"),
+        ("httpcore", logging.DEBUG, "connection.close.complete"),
+    ]
+
+
 @pytest.mark.anyio
 async def test_connection_pool_with_http_exception():
     """
@@ -201,8 +277,12 @@ async def test_connection_pool_with_connect_exception():
 
     class FailedConnectBackend(AsyncMockBackend):
         async def connect_tcp(
-            self, host: str, port: int, timeout: float = None, local_address: str = None
-        ):
+            self,
+            host: str,
+            port: int,
+            timeout: Optional[float] = None,
+            local_address: Optional[str] = None,
+        ) -> AsyncNetworkStream:
             raise ConnectError("Could not connect")
 
     network_backend = FailedConnectBackend([])
@@ -325,10 +405,12 @@ async def test_connection_pool_concurrency():
             for domain in ["a.com", "b.com", "c.com", "d.com", "e.com"]:
                 nursery.start_soon(fetch, pool, domain, info_list)
 
-        # Check that each time we inspect the connection pool, only a
-        # single connection was established.
         for item in info_list:
+            # Check that each time we inspected the connection pool, only a
+            # single connection was established at any one time.
             assert len(item) == 1
+            # Each connection was to a different host, and only sent a single
+            # request on that connection.
             assert item[0] in [
                 "<AsyncHTTPConnection ['http://a.com:80', HTTP/1.1, ACTIVE, Request Count: 1]>",
                 "<AsyncHTTPConnection ['http://b.com:80', HTTP/1.1, ACTIVE, Request Count: 1]>",
@@ -349,6 +431,7 @@ async def test_connection_pool_concurrency_same_domain_closing():
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
             b"Content-Length: 13\r\n",
+            b"Connection: close\r\n",
             b"\r\n",
             b"Hello, world!",
         ]
@@ -368,17 +451,15 @@ async def test_connection_pool_concurrency_same_domain_closing():
             for domain in ["a.com", "a.com", "a.com", "a.com", "a.com"]:
                 nursery.start_soon(fetch, pool, domain, info_list)
 
-        # Check that each time we inspect the connection pool, only a
-        # single connection was established.
         for item in info_list:
+            # Check that each time we inspected the connection pool, only a
+            # single connection was established at any one time.
             assert len(item) == 1
-            assert item[0] in [
-                "<AsyncHTTPConnection ['https://a.com:443', HTTP/1.1, ACTIVE, Request Count: 1]>",
-                "<AsyncHTTPConnection ['https://a.com:443', HTTP/1.1, ACTIVE, Request Count: 2]>",
-                "<AsyncHTTPConnection ['https://a.com:443', HTTP/1.1, ACTIVE, Request Count: 3]>",
-                "<AsyncHTTPConnection ['https://a.com:443', HTTP/1.1, ACTIVE, Request Count: 4]>",
-                "<AsyncHTTPConnection ['https://a.com:443', HTTP/1.1, ACTIVE, Request Count: 5]>",
-            ]
+            # Only a single request was sent on each connection.
+            assert (
+                item[0]
+                == "<AsyncHTTPConnection ['https://a.com:443', HTTP/1.1, ACTIVE, Request Count: 1]>"
+            )
 
 
 @pytest.mark.trio
@@ -412,10 +493,11 @@ async def test_connection_pool_concurrency_same_domain_keepalive():
             for domain in ["a.com", "a.com", "a.com", "a.com", "a.com"]:
                 nursery.start_soon(fetch, pool, domain, info_list)
 
-        # Check that each time we inspect the connection pool, only a
-        # single connection was established.
         for item in info_list:
+            # Check that each time we inspected the connection pool, only a
+            # single connection was established at any one time.
             assert len(item) == 1
+            # The connection sent multiple requests.
             assert item[0] in [
                 "<AsyncHTTPConnection ['https://a.com:443', HTTP/1.1, ACTIVE, Request Count: 1]>",
                 "<AsyncHTTPConnection ['https://a.com:443', HTTP/1.1, ACTIVE, Request Count: 2]>",
@@ -433,3 +515,91 @@ async def test_unsupported_protocol():
 
         with pytest.raises(UnsupportedProtocol):
             await pool.request("GET", "://www.example.com/")
+
+
+@pytest.mark.anyio
+async def test_connection_pool_closed_while_request_in_flight():
+    """
+    Closing a connection pool while a request/response is still in-flight
+    should raise an error.
+    """
+    network_backend = AsyncMockBackend(
+        [
+            b"HTTP/1.1 200 OK\r\n",
+            b"Content-Type: plain/text\r\n",
+            b"Content-Length: 13\r\n",
+            b"\r\n",
+            b"Hello, world!",
+        ]
+    )
+
+    async with AsyncConnectionPool(
+        network_backend=network_backend,
+    ) as pool:
+        # Send a request, and then close the connection pool while the
+        # response has not yet been streamed.
+        async with pool.stream("GET", "https://example.com/") as response:
+            await pool.aclose()
+            with pytest.raises(ReadError):
+                await response.aread()
+
+
+@pytest.mark.anyio
+async def test_connection_pool_timeout():
+    """
+    Ensure that exceeding max_connections can cause a request to timeout.
+    """
+    network_backend = AsyncMockBackend(
+        [
+            b"HTTP/1.1 200 OK\r\n",
+            b"Content-Type: plain/text\r\n",
+            b"Content-Length: 13\r\n",
+            b"\r\n",
+            b"Hello, world!",
+        ]
+    )
+
+    async with AsyncConnectionPool(
+        network_backend=network_backend, max_connections=1
+    ) as pool:
+        # Send a request to a pool that is configured to only support a single
+        # connection, and then ensure that a second concurrent request
+        # fails with a timeout.
+        async with pool.stream("GET", "https://example.com/"):
+            with pytest.raises(PoolTimeout):
+                extensions = {"timeout": {"pool": 0.0001}}
+                await pool.request("GET", "https://example.com/", extensions=extensions)
+
+
+@pytest.mark.anyio
+async def test_http11_upgrade_connection():
+    """
+    HTTP "101 Switching Protocols" indicates an upgraded connection.
+
+    We should return the response, so that the network stream
+    may be used for the upgraded connection.
+
+    https://httpwg.org/specs/rfc9110.html#status.101
+    https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/101
+    """
+    network_backend = AsyncMockBackend(
+        [
+            b"HTTP/1.1 101 Switching Protocols\r\n",
+            b"Connection: upgrade\r\n",
+            b"Upgrade: custom\r\n",
+            b"\r\n",
+            b"...",
+        ]
+    )
+    async with AsyncConnectionPool(
+        network_backend=network_backend, max_connections=1
+    ) as pool:
+        async with pool.stream(
+            "GET",
+            "wss://example.com/",
+            headers={"Connection": "upgrade", "Upgrade": "custom"},
+        ) as response:
+            assert response.status == 101
+            network_stream = response.extensions["network_stream"]
+            content = await network_stream.read(max_bytes=1024)
+            assert content == b"..."
